import flask
from requests.exceptions import HTTPError

from canonicalwebteam.discourse_docs.models import (
    NavigationParseError,
    RedirectFoundError,
)


def build_blueprint(url_prefix, model, document_template):

    blueprint = flask.Blueprint("discourse_docs", __name__)

    def redirect_handler(path):
        if url_prefix != "/":
            return flask.redirect(url_prefix + path)
        else:
            return flask.redirect(path)

    @blueprint.route("/")
    def homepage():
        """
        Redirect to the frontpage topic
        """
        frontpage, nav_html = model.parse_frontpage()
        return redirect_handler(frontpage["path"])

    @blueprint.route("/t/<path:path>")
    def redirect(path):
        return redirect_handler("/" + path)

    @blueprint.route("/<path:path>")
    def document(path):
        try:
            document, nav_html = model.get_document(path)
        except RedirectFoundError as redirect_error:
            return flask.redirect(redirect_error.redirect_path)
        except HTTPError as http_error:
            flask.abort(http_error.response.status_code)
        except NavigationParseError as nav_error:
            document = nav_error.document
            nav_html = f"<p>{str(nav_error)}</p>"

        return flask.render_template(
            document_template,
            title=document["title"],
            body_html=document["body_html"],
            forum_link=document["forum_link"],
            updated=document["updated"],
            nav_html=nav_html,
        )

    return blueprint



import re
from urllib.parse import urlparse

from jinja2 import Template
from requests.exceptions import HTTPError

import dateutil.parser
import humanize
from bs4 import BeautifulSoup
from canonicalwebteam.http import CachedSession


class RedirectFoundError(HTTPError):
    """
    If we encounter redirects from Discourse, we need to take action
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        url_parts = urlparse(self.response.headers["Location"])
        self.redirect_path = re.sub("/t(/.*).json", r"\1", url_parts.path)


class NotInCategoryError(HTTPError):
    """
    If we encounter redirects from Discourse, we need to take action
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        url_parts = urlparse(self.response.headers["Location"])
        self.redirect_path = re.sub("/t(/.*).json", r"\1", url_parts.path)


class NavigationParseError(Exception):
    """
    Indicates a failure to extract the navigation from
    the frontpage content
    """

    def __init__(self, document, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.document = document


# Private helper functions
# ===


def _process_html(html):
    """
    Post-process the HTML output from Discourse to
    remove 'NOTE TO EDITORS' sections.

    We expect these sections to be of the HTML format:

    <aside class="quote no-group">
      <blockquote>
        <p>
          <img title=":construction:" class="emoji" ...>
          <strong>NOTE TO EDITORS</strong>
          <img title=":construction:" class="emoji" ...>
        </p>
        <p> ... </p>
      </blockquote>
    </aside>
    """

    soup = BeautifulSoup(html, features="html.parser")
    notes_to_editors_text = soup.find_all(text="NOTE TO EDITORS")

    soup = _replace_notifications(soup)

    for text in notes_to_editors_text:
        # If this section is of the expected HTML format,
        # we should find the <aside> container 4 levels up from
        # the "NOTE TO EDITORS" text
        container = text.parent.parent.parent.parent

        if container.name == "aside" and "quote" in container.attrs["class"]:
            container.decompose()

    return str(soup)


def _replace_notifications(soup):
    """
    Given some BeautifulSoup of a document,
    replace blockquotes with the appropriate notification markup
    """
    notification_html = (
        "<div class='{{ notification_class }}'>"
        "<div class='p-notification__response'>"
        "{{ contents | safe }}"
        "</div></div>"
    )

    notification_template = Template(notification_html)
    for note_string in soup.findAll(text=re.compile("ⓘ ")):
        first_paragraph = note_string.parent
        blockquote = first_paragraph.parent
        last_paragraph = blockquote.findChildren(recursive=False)[-1]

        if first_paragraph.name == "p" and blockquote.name == "blockquote":
            # Remove extra padding/margin
            first_paragraph.attrs["class"] = "u-no-padding--top"
            if last_paragraph.name == "p":
                if "class" in last_paragraph.attrs:
                    last_paragraph.attrs["class"] += " u-no-margin--bottom"
                else:
                    last_paragraph.attrs["class"] = "u-no-margin--bottom"

            # Remove control emoji
            notification_html = blockquote.encode_contents().decode("utf-8")
            notification_html = re.sub(
                r"^\n<p([^>]*)>ⓘ +", r"<p\1>", notification_html
            )

            notification = notification_template.render(
                notification_class="p-notification", contents=notification_html
            )
            blockquote.replace_with(
                BeautifulSoup(notification, features="html.parser")
            )

    for warning in soup.findAll("img", title=":warning:"):
        first_paragraph = warning.parent
        blockquote = first_paragraph.parent
        last_paragraph = blockquote.findChildren(recursive=False)[-1]

        if first_paragraph.name == "p" and blockquote.name == "blockquote":
            warning.decompose()

            # Remove extra padding/margin
            first_paragraph.attrs["class"] = "u-no-padding--top"
            if last_paragraph.name == "p":
                if "class" in last_paragraph.attrs:
                    last_paragraph.attrs["class"] += " u-no-margin--bottom"
                else:
                    last_paragraph.attrs["class"] = "u-no-margin--bottom"

            notification = notification_template.render(
                notification_class="p-notification--caution",
                contents=blockquote.encode_contents().decode("utf-8"),
            )
            blockquote.replace_with(
                BeautifulSoup(notification, features="html.parser")
            )

    return soup


class DiscourseAPI:
    """
    A basic model class for retrieving Documentation content
    from a Discourse installation through the API
    """

    def __init__(
        self,
        base_url,
        frontpage_id,
        category_id,
        session=CachedSession(fallback_cache_duration=300),
    ):
        """
        @param base_url: The Discourse URL (e.g. https://discourse.example.com)
        @param frontpage_id: The ID of the frontpage topic in Discourse.
                            This topic should also contain the navigation.
        """

        self.base_url = base_url.rstrip("/")
        self.frontpage_id = frontpage_id
        self.category_id = category_id
        self.session = session

    def __del__(self):
        self.session.close()

    def get_document(self, path):
        """
        Retrieve and return relevant data about a document:
        - Title
        - HTML content
        - Navigation content
        """

        parse_error = None

        try:
            frontpage, nav_html = self.parse_frontpage()
        except NavigationParseError as err:
            parse_error = err
            frontpage = parse_error.document

        if f"{self.base_url}/t/{path}" == frontpage["forum_link"]:
            document = frontpage
        else:
            document = self._parse_document_topic(self._get_topic(path))

        if parse_error:
            parse_error.document = document
            raise parse_error

        return document, nav_html

    def parse_frontpage(self):
        """
        Parse the frontpage document topic to extract the Navigation markup
        from it
        """

        # Get topic data
        frontpage_topic = self._get_topic(self.frontpage_id)
        frontpage_document = self._parse_document_topic(frontpage_topic)

        # Split HTML into nav and body
        soup = BeautifulSoup(
            frontpage_document["body_html"], features="html.parser"
        )
        splitpoint = soup.find(re.compile("^h[1-6]$"), text="Content")

        if splitpoint:
            body_elements = splitpoint.fetchPreviousSiblings()
            frontpage_document["body_html"] = "".join(
                map(str, reversed(body_elements))
            )

            nav_elements = splitpoint.fetchNextSiblings()
            nav_html = "".join(map(str, nav_elements))
            nav_soup = BeautifulSoup(nav_html, features="html.parser")

            # Convert links to the form needed in this site
            for link in nav_soup.find_all("a"):
                if "href" in link.attrs:
                    url = link.attrs["href"]
                    link_match = f"^({self.base_url})?/t(/.*)$"
                    link.attrs["href"] = re.sub(link_match, r"\2", url)
        else:
            raise NavigationParseError(
                frontpage_document,
                "Error: Failed to parse navigation from "
                + frontpage_document["forum_link"]
                + ". Please check the format.",
            )

        return frontpage_document, str(nav_soup)

    # Private helper methods
    # ===

    def _parse_document_topic(self, topic):
        """
        Parse a topic object retrieve from Discourse
        and return document data:
        - title: The title
        - body_html: The HTML content of the initial topic post
                     (with some post-processing)
        - updated: A human-readable data, relative to now
                   (e.g. "3 days ago")
        - forum_link: The link to the original forum post
        """

        updated_datetime = dateutil.parser.parse(
            topic["post_stream"]["posts"][0]["updated_at"]
        )

        return {
            "title": topic["title"],
            "body_html": _process_html(
                topic["post_stream"]["posts"][0]["cooked"]
            ),
            "updated": humanize.naturaltime(
                updated_datetime.replace(tzinfo=None)
            ),
            "forum_link": f"{self.base_url}/t/{topic['slug']}/{topic['id']}",
            "path": f"/{topic['slug']}/{topic['id']}",
        }

    def _get_topic(self, path):
        """
        Retrieve topic object by path
        """

        response = self.session.get(
            f"{self.base_url}/t/{path}.json", allow_redirects=False
        )
        response.raise_for_status()

        if response.status_code >= 300:
            raise RedirectFoundError(response=response)

        topic = response.json()

        # If topic not in category, raise a 404
        if topic["category_id"] != self.category_id:
            error = HTTPError(f"Topic not in category {self.category_id}")
            response.status_code = 404
            error.response = response
            raise error

        return response.json()



import unittest

from requests import Session
from requests.exceptions import HTTPError

import requests_mock
from canonicalwebteam.discourse_docs import models
from tests.fixtures.snapcraft_forum_mock import register_uris


class TestDiscourseApi(unittest.TestCase):
    def setUp(self):
        self.mock_session = Session()
        snapcraft_mock_adapter = requests_mock.Adapter()
        self.mock_session.mount("https://", snapcraft_mock_adapter)
        register_uris(snapcraft_mock_adapter)

        self.discourse = models.DiscourseApi(
            base_url="https://forum.snapcraft.io",
            frontpage_id=3781,
            category_id=15,  # The "doc" category
            session=self.mock_session,
        )

    def test_get_document_redirects(self):
        """
        Check that if a document topic has been renamed in Discourse,
        the DiscourseApi model will return a RedirectFoundError,
        which will contain the "redirect_path".

        The intention is then for the client app to handle the redirect
        error, e.g. by sending a redirect response to the user.
        """

        # Catch the error
        with self.assertRaises(models.RedirectFoundError) as context:
            self.discourse.get_document("configuration-options/87")

        # Check the redirect_path points to the right place
        self.assertEqual("/system-options/87", context.exception.redirect_path)

    def test_get_document_not_found(self):
        """
        If a document topic doesn't exist in Discourse,
        the DiscourseApi model should return a 404 requests HTTPError,
        which the client can then handle if they wish.
        """

        # Catch the error
        with self.assertRaises(HTTPError) as context:
            self.discourse.get_document("fictional-document")

        # Check it communicates the appropriate HTTP status
        self.assertEqual(404, context.exception.response.status_code)

    def test_get_document_not_in_category(self):
        """
        Check requesting a topic not in the selected category
        leads to a 404
        """

        # Catch error
        with self.assertRaises(HTTPError) as context:
            self.discourse.get_document(
                "script-to-remove-expired-kerberos-caches-for-"
                "enhanced-stability/10096"
            )

        # Check it's a 404
        self.assertEqual(404, context.exception.response.status_code)

    def test_get_document_broken_nav(self):
        """
        Check that, when we request a document that is different
        from the frontpage document, but the frontpage is set
        to a topic which doesn't contain a navigation section
        in the expected format,  that it will raise a NavigationParseError,
        but still return the requested document
        as NavigationParseError.document
        """

        # Set up a Discourse pointing at a badly formatted frontpage
        broken_discourse = models.DiscourseApi(
            base_url="https://forum.snapcraft.io",
            frontpage_id=3876,
            category_id=15,  # The "doc" category
            session=self.mock_session,
        )

        # Catch the NavigationParseError
        with self.assertRaises(models.NavigationParseError) as context:
            document, nav_html = broken_discourse.get_document(
                "documentation-outline/3781"
            )

        # Retrieve the document from the error
        nav_error = context.exception
        doc = nav_error.document

        # Check the shape of the document content
        self.assertEqual(doc["title"], "Documentation outline")
        self.assertTrue(bool(doc.get("updated")))
        self.assertTrue("the experimental snap" in doc["body_html"])
        self.assertEqual(
            doc["forum_link"],
            "https://forum.snapcraft.io/t/documentation-outline/3781",
        )

    def test_get_document_topic(self):
        """
        Check that DiscourseApi.get_document is able to parse
        a basic topic thread (as opposed to a wiki thread)
        and build an appropriately-shaped document dictionary.
        """

        # Get the document and navigation
        topic_doc, topic_nav_html = self.discourse.get_document(
            "documentation-outline/3781"
        )

        # Check the shape of the document content
        self.assertEqual(topic_doc["title"], "Documentation outline")
        self.assertTrue(bool(topic_doc.get("updated")))
        self.assertTrue("the experimental snap" in topic_doc["body_html"])
        self.assertEqual(
            topic_doc["forum_link"],
            "https://forum.snapcraft.io/t/documentation-outline/3781",
        )

        # Check the shape of the navigation content
        self.assertTrue("<h3>Publishing</h3>" in topic_nav_html)
        self.assertTrue(
            '<a href="/the-maven-plugin/4282">Maven</a>' in topic_nav_html
        )

    def test_get_document_wiki(self):
        """
        Check that DiscourseApi.get_document is able to parse
        a basic wiki thread (as opposed to a basic topic thread)
        and build an appropriately-shaped document dictionary.
        """

        # Get the document and navigation
        wiki_doc, wiki_nav_html = self.discourse.get_document(
            "getting-started/3876"
        )

        # Check the shape of the document content
        self.assertEqual(wiki_doc["title"], "Getting started")
        self.assertTrue(bool(wiki_doc.get("updated")))
        self.assertFalse("NOTE TO EDITORS" in wiki_doc["body_html"])
        self.assertTrue("<p>The following" in wiki_doc["body_html"])
        self.assertEqual(
            wiki_doc["forum_link"],
            "https://forum.snapcraft.io/t/getting-started/3876",
        )

        # Check the shape of the navigation content
        self.assertTrue("<h3>Publishing</h3>" in wiki_nav_html)
        self.assertTrue(
            '<a href="/the-maven-plugin/4282">Maven</a>' in wiki_nav_html
        )

    def test_parse_frontpage_success(self):
        """
        Check that DiscourseApi is able to retrieve the frontpage topic
        and successfully split the basic document body HTML
        and the content into the navigation HTML
        """

        # Get frontpage and nav
        frontpage, nav_html = self.discourse.parse_frontpage()

        # Check the shape of the document content
        self.assertFalse("<h1>Content</h1>" in frontpage["body_html"])
        self.assertFalse("<h3>Publishing</h3>" in frontpage["body_html"])
        self.assertTrue("Choose the topic" in frontpage["body_html"])
        self.assertEqual("Documentation outline", frontpage["title"])
        self.assertEqual(
            "https://forum.snapcraft.io/t/documentation-outline/3781",
            frontpage["forum_link"],
        )

        # Check the shape of the navigation
        self.assertTrue("<h3>Publishing</h3>" in nav_html)
        self.assertTrue(
            '<a href="/the-maven-plugin/4282">Maven</a>' in nav_html
        )

    def test_parse_frontpage_navigation_error(self):
        """
        Check that if the frontpage is set to a topic which doesn't
        contain a navigation section in the expected format,
        that it will raise a NavigationParseError, but still
        returning the original document as NavigationParseError.document
        """

        # Set up a Discourse pointing at a badly formatted frontpage
        broken_discourse = models.DiscourseApi(
            base_url="https://forum.snapcraft.io",
            frontpage_id=3876,
            category_id=15,  # The "doc" category
            session=self.mock_session,
        )

        # Catch the NavigationParseError
        with self.assertRaises(models.NavigationParseError) as context:
            frontpage, nav_html = broken_discourse.parse_frontpage()

        # Retrieve the document from the error
        nav_error = context.exception
        doc = nav_error.document

        # Check the document still looks as expected
        self.assertEqual(doc["title"], "Getting started")
        self.assertFalse("NOTE TO EDITORS" in doc["body_html"])
        self.assertTrue("<p>The following sections" in doc["body_html"])
        self.assertTrue(doc["forum_link"] in str(context.exception))

    def test_notifications(self):
        """
        Check that topics with notifications get converted properly
        to Vanilla notification markup
        """

        # Get the document and navigation
        doc_with_note, nav_html = self.discourse.get_document(
            "choosing-a-security-model/6847"
        )
        doc_with_warning, nav_html = self.discourse.get_document(
            "service-management/3965"
        )

        # Check the shape of notification
        note_li = (
            "<li>Open arbitrary filesystem paths outside of "
            "<code>$HOME</code> and /media</li>"
        )
        note_end_p = (
            '<p class="u-no-margin--bottom">In these cases, '
            '<a href="/t/snap-confinement/6233">“classic” confinement</a> '
            "may be an option.</p>"
        )

        self.assertTrue('class="p-notification"' in doc_with_note["body_html"])
        self.assertTrue(note_li in doc_with_note["body_html"])
        self.assertTrue(note_end_p in doc_with_note["body_html"])
        self.assertTrue("ⓘ" not in doc_with_note["body_html"])

        # Check the shape of the warning
        warning_start_p = (
            '<p class="u-no-padding--top u-no-margin--bottom"> Stopping snap '
            "services manually may cause the snap to malfunction. For "
            "temporarily disabling a snap consider using the <em>enable</em> "
            "and <em>disable</em> commands instead.</p>"
        )
        self.assertTrue(
            'class="p-notification--caution"' in doc_with_warning["body_html"]
        )
        self.assertTrue(warning_start_p in doc_with_warning["body_html"])


if __name__ == "__main__":
    unittest.main()
